//---------------------------------------------------------------------------
// 7A_Floyd_shortest_path.cpp 
//---------------------------------------------------------------------------
#ifdef _WIN32
	#pragma warning(disable:4996)
	#include <windows.h>
	#include <conio.h>
	#include <direct.h>
#elif (defined __linux__) || (defined _AIX) || (defined __APPLE__)
	#include <stdlib.h>
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <unistd.h>
	typedef char _TCHAR;
	#define _tmain main
#endif

#include <mpi.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
using namespace std;
//---------------------------------------------------------------------------
#define MIN(A,b)                 ((A)<(b)?(A):(b))		
// napr. nech mame 3 procesy a 6 vrcholov grafu ciest
// (cislovanie riadkov matice ciest zacina od 0)
#define BLOCK_LOW(rank,size,n)   ((rank)*(n)/(size))
// napr. pre stredny pruh dostaneme 1 * 6 / 3 = 2
#define BLOCK_HIGH(rank,size,n)  (BLOCK_LOW((rank)+1,size,n)-1) 
// 2 * 6 / 3 - 1 = 2 * 2 - 1 = 3
#define BLOCK_SIZE(rank,size,n)  (BLOCK_HIGH(rank,size,n)-BLOCK_LOW(rank,size,n)+1) 
// 3 - 2 + 1 = 2

#define BLOCK_OWNER(j,size,n)    (((size)*((j)+1)-1)/(n)) 
// napr. nech j = 1, potom (3 * (1 + 1 -1)) / 6 = 3 / 6 = 0 (celociselne delenie)
// nech j = 2. potom (3 * (2 + 1 -1)) / 6 = 6 / 6 = 1
#define MAXLINE 255													 
// nech j = 4. potom (3 * (4 + 1 -1)) / 6 = 12 / 6 = 2
// nech j = 5. potom (3 * (5 + 1 -1)) / 6 = 15 / 6 = 2 (celociselne delenie)
//--------------------------------------------------------
void compute_shortest_paths(int, int, int[][6], int);
void GetNamesOfIOFiles(const char[], char[]);
//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	int A[6][6];   /* Doubly-subscripted array */
	int i, j, k;
	int size;	   /* Number of processes */
	int rank;	   /* Process rank */
	int m=0;	      /* Rows in matrix */
	int n=0;	      /* Columns in matrix */
	int tmp;
	FILE* in;
	char nameoffile[MAXLINE];
	double my_time;
	MPI_Datatype my_type1;

	MPI_Init(&argc, &argv);
	MPI_Barrier(MPI_COMM_WORLD);
	my_time = -MPI_Wtime();

	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	if (6 % size != 0) {
		if (rank == 0)
			printf("\nThe number of processes should be 1, 2, 3 or 6!\n");
		goto end;
	}
	if (rank == 0)
		cout << "\nThere are " << size << " processes.\n";

	GetNamesOfIOFiles("distances.txt", nameoffile);
	if (rank == 0) {
		in = fopen(nameoffile, "r");
		tmp = fscanf(in, "%i %i", &m, &n);
		for (i = 0; i < m; i++)
			for (j = 0; j < n; j++)
				tmp = fscanf(in, "%i", &A[i][j]);
		fclose(in);

		if (m != n) {
			cout << "Matrix must be square\n";
			goto end;
		}

		printf("\nThe original graph of distances was:\n\n");
		for (i = 0; i < m; i++) {
			for (j = 0; j < n; j++)
				printf("%4i", A[i][j]);
			printf("\n");
		}
		printf("\n");
	}
	MPI_Bcast(&m, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

	MPI_Type_vector(BLOCK_SIZE(rank, size, n) * 6, 1, 1, MPI_INT, &my_type1);
	MPI_Type_commit(&my_type1);

	if (rank == 0)
		for (k = 1; k < size; k++) // posielame vodorovne pruhy matice A
			MPI_Ssend(&(A[BLOCK_LOW(k, size, n)][0]), 1, my_type1, k, k, 
				MPI_COMM_WORLD);
	else
		MPI_Recv(&(A[0][0]), 1, my_type1, 0, rank, MPI_COMM_WORLD, 
			MPI_STATUSES_IGNORE);

	compute_shortest_paths(rank, size, A, n);

	if (rank > 0)
		MPI_Ssend(&(A[0][0]), 1, my_type1, 0, rank, MPI_COMM_WORLD);
	if (rank == 0)
		for (k = 1; k < size; k++)
			MPI_Recv(&(A[BLOCK_LOW(k, size, n)][0]), 1, my_type1, k, k, MPI_COMM_WORLD, 
				MPI_STATUSES_IGNORE);

	if (rank == 0) {
		printf("\nThe graph of the shortest distances is:\n\n");
		for (i = 0; i < m; i++) {
			for (j = 0; j < n; j++)
				printf("%4i", A[i][j]);
			printf("\n");
		}
		printf("\n");
	}

end:
	my_time += MPI_Wtime();
	if (rank == 0)
		printf("Time = %10.6f s\n", my_time);

	fflush(stdout);
	MPI_Finalize();

	return 0;
}
//---------------------------------------------------------------------------
void compute_shortest_paths(int rank, int size, int A[][6], int n)
{
	int  i, j, k;
	int  offset;     /* Local index of broadcast row */
	int  root;       /* Process controlling row to be beast */
	int  tmp[6];     /* Holds the broadcast row */
	
	for (j = 0; j < n; j++)
		tmp[0] = 0;

	for (k = 0; k < n; k++) {
		root = BLOCK_OWNER(k, size, n);
		if (root == rank) {
			offset = k - BLOCK_LOW(rank, size, n);  
			// napr. ak rank = 1 a size = 3, potom k nadobuda hodnoty 0 a 1
			for (j = 0; j < n; j++)						 
				tmp[j] = A[offset][j]; 
			// lebo kopie pruhov matice A sme ukladali na &A[0][0] 
			// tmp je kopia riadku matice A
		}
		// MPI_Bcast synchronizuje procesy ako kazde collective
		// communication, t.j., vsetky procesy budu mat rovnake k
		MPI_Bcast(tmp, n, MPI_INT, root, MPI_COMM_WORLD); 
		for (i = 0; i < BLOCK_SIZE(rank, size, n); i++)   
			for (j = 0; j < n; j++)                         
				A[i][j] = MIN(A[i][j], A[i][k] + tmp[j]);   
		// hladame minimum z A[i][j] a vsetkych moznych A[i][k] + A[k][j]
		// A[i][j] = MIN(A[i][j],A[i][k]+A[k][j]) !!!!!!!!!!!!!!!
	}                                                    
}
//---------------------------------------------------------------------------
void GetNamesOfIOFiles(const char name_of_input_file[], char path_to_input_file[])
{
	char current_path[MAXLINE];
	current_path[0] = '\0';

#ifdef _WIN32
	TCHAR exePath[MAXLINE];

	HMODULE hModule = GetModuleHandle(NULL);
	if (hModule != NULL) {
		if (!GetModuleFileName(hModule, exePath, MAXLINE)) {
			cout << "Nepodarila sa zistit cesta k exe-suboru.\n";
			exit(1);
		}
	}
	else {
		cout << "Module handle is NULL.\n" << endl;
		exit(1);
	}

	int iii;
	bool flag = false;
	for (iii = (int)wcslen(exePath); iii >= 0; iii--) {
		if (!flag && exePath[iii] == '\\') {
			current_path[iii + 1] = '\0';
			flag = true;
		}
		if (flag)
			current_path[iii] = (char)exePath[iii];
	}
#elif (defined __linux__) || (defined __APPLE__)
	unsigned iii;
	char line[MAXLINE];
	FILE* fp;
	if ((fp = popen("/bin/pwd", "r")) == NULL) {
		perror("popen error");
		exit(1);
	}
	if (fgets(line, MAXLINE, fp) == NULL) {
		perror("fgets error");
		exit(1);
	}
	pclose(fp);

	iii = 0;
	while (line[iii] != '\r' && line[iii] != '\n') {
		current_path[iii] = line[iii];
		iii++;
	}
	current_path[iii] = '\0';
#elif (defined _AIX)
	unsigned iii;
	char line[MAXLINE];
	FILE* fp;
	if ((fp = popen("user/bin/pwd", "r")) == NULL) {
		perror("popen error");
		exit(1);
	}
	if (fgets(line, MAXLINE, fp) == NULL) {
		perror("fgets error");
		exit(1);
	}
	pclose(fp);

	iii = 0;
	while (line[iii] != '\r' && line[iii] != '\n') {
		current_path[iii] = line[iii];
		iii++;
	}
	current_path[iii] = '\0';
#endif

	path_to_input_file[0] = '\0';
	strcat(path_to_input_file, current_path);
#if (defined __linux__) || (defined _AIX) || (defined __APPLE__)
	strcat(path_to_input_file, "/inputs/");
#elif (defined _WIN32)
	strcat(path_to_input_file, "inputs\\");
#endif
	strcat(path_to_input_file, name_of_input_file);
}